/** 
 * @file MainForm.cs
 * @author Yarrago
 * @note Copyright  Yarrago (yarrago.com) 2009 - 2010
 * @version 1.0.5
 * @date Created: 23/11/2009 ?
 * @date Modified: 29/11/2010
 * 
 * @note Modified by Yarrago on 13/04/2010. Description: Added thumbnail preview mode.
 * @note Modified by Yarrago on 15/04/2010. Description: Fixed a bug that changing the category would not work.
 * @note Modified by Yarrago on 18/04/2010. Description: Added info status label, added length to file name, added pause auto timer button, increased auto timer time, added open current folder in explorer, added warning when file type changes.
 * @note Modified by Yarrago on 01/05/2010. Description: Added drag and drop functionality to the main image and added batch processing from the commandline (and sent to remote instances).
 * @note Modified by Yarrago on 17/06/2010. Description: Added a button to load the main image in system default application.
 * @note Modified by Yarrago on 29/11/2010. Description: Implemented recycler so now files are deleted to the recycling bin.
 * @note Modified by Name on --/--/2010. Description:
 * 
 * 
 * 
 * 
 * 
 * License: "Modified BSD License" 3-clause license.
 * 
 * Copyright (c) 2010, Yarrago (yarrago.com)
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without modification, are 
 * permitted provided that the following conditions are met:
 *   - Redistributions of source code must retain the above copyright notice, this list of 
 *     conditions and the following disclaimer.
 *   - Redistributions in binary form must reproduce the above copyright notice, this list of 
 *     conditions and the following disclaimer in the documentation and/or other materials provided
 *     with the distribution.
 *   - Neither the name of the yarrago.com nor the names of its contributors may be used to endorse
 *     or promote products derived from this software without specific prior written permission.
 *   
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR 
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR 
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR 
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR 
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 */

using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using System.Runtime.InteropServices;
using System.Diagnostics;

namespace ScanHelper
{
    public partial class MainForm : Form
    {
        private const uint WM_COPYDATA = 0x4A;

        // Declare CopyDataStruct for use with SendMessage.
        // Note: Not sure that this declaration is accurate (it works, but may cause problem ie 32bit / 64bit if it is not correct).
        [StructLayout(LayoutKind.Sequential)]
        private struct CopyDataStruct
        {
            public UInt32 dwData;   // ULONG_PTR dwData;    // Specifies data to be passed to the receiving application.
            public UInt32 cbData;   // DWORD cbData;        // Specifies the size, in bytes, of the data pointed to by the lpData member. 
            public IntPtr lpData;   // PVOID lpData;        // Pointer to data to be passed to the receiving application. This member can be NULL.          
        }

        // Import the unmanaged SetForegroundWindow() function declaration.
        [DllImport("user32.dll", CharSet = CharSet.Auto)]
        private static extern bool SetForegroundWindow(IntPtr hWnd);


        // Import the unmanaged SendMessage() function declaration.
        // Note: Not sure that this declaration is accurate (it works, but may cause problem ie 32bit / 64bit if it is not correct).
        [DllImport("user32.dll", CharSet = CharSet.Auto, SetLastError = true)]
        private static extern IntPtr SendMessage(IntPtr hWnd,
                                        UInt32 Msg,
                                        IntPtr wParam,
                                        IntPtr lParam);

        #if STRING_METHOD_1
        // Note: Not sure that this declaration is accurate (it works, but may cause problem ie 32bit / 64bit if it is not correct).
        [DllImport("kernel32.dll", CharSet = CharSet.Auto)]           // Method 1: In method 1 the number of bytes that are put on the heap arn't known so we need to use this method to find out.
        private static extern UIntPtr GlobalSize(IntPtr hMem);
        #endif





        private const bool actuallyRename = true;                           // This constant is used in debugging to turn on or off whether the file is actually renamed or not.

        // Constant Values Used To Store Information To File.
        private const string FileCategoryDelimiter = ": ";
        private const string FileConfirmRename = "ConfirmRename";
        private const string FileUseThumbnailMode = "UseThumbnailMode";
        private const string FileFocusApplication = "FocusApplication";
        private const string FileAutoRename = "AutoRename";
        private const string FileAutoRenameAllowedAutomated = "AutoRenameAutomated";
        private const string FileAutoRenameAllowedManual = "AutoRenameManual";
        private const string FileBaseDirectory = "BaseDirectory";
        private const string FileLastProfile = "LastProfile";
        private const string FileProfileDelimiter = "-----";
        private const string FileProfileCategory = "ProfileCategory";
        private const string FileProfileName = "ProfileName";
        private const string FileProfileDirectory = "ProfileDirectory";
        private const string FileProfileFileName = "ProfileFileName";
        private const string FileProfileMask = "ProfileMask";
        private const string FileProfileAutoUpdate = "ProfileAutoUpdate";

        private const string filenameConjoiner = "\"";                      // The character used to wrap filenames with when they are conjoined for passing between instances.

        private DateTime startTime = DateTime.Now;                          // Stores the open time of the application used for statistics.
        private int filesOpened = 0;                                        // Counts the number of files that have been opened since the program loaded.
        private const int ShortAverageDuration = 10;                        // The number of minutes to create a short average over.
        private int[] filesOpenedMinute = new int[ShortAverageDuration];    // Bins to hold how many files were opened in the last few minutes.
        private long statsLastMinute = 0;                                   // The last minute the bins holding the minutes were rotated.

        private bool autoUpdateInfoStatus = true;                           // Whether to auto update the info status label.
        private int currentInfoStatus = 0;                                  // The current status to be displayed by the infro status label.
        private const int TotalInfoStatus = 4;                              // The total number of info status labels.

        private const bool WarnImageTypeChanged = true;                     // Whether to warn if the image type changes.
        private string lastImageFormat = "";                                // The format of the last image loaded by the program.
        bool lastImageFormatValid = false;                                  // Whether the lastImageFormat is valid I.E. Has been set.

        private const int imageCacheSize = 5;                               // The number of images to store in the internal cache.
        private ImageCache[] imageCache = new ImageCache[imageCacheSize];   // The actual image cache.

        private bool allowSaveSettings = false;                             // This stops the dummy inastances from overwriting the settings file when multiple instances are open.
        private DateTime fileLastModifiedTime = DateTime.Now;               // The last time the settings file was modified when it was last accessed.
        private string settingsPath = System.IO.Path.Combine(Environment.GetFolderPath(Environment.SpecialFolder.MyDocuments), @"ScanHelperSettings.txt");
        string dumpFolder = System.IO.Path.Combine(Environment.GetFolderPath(Environment.SpecialFolder.Desktop), @"Scan Helper Deleted Images");
        
        private string focusProcessName = "mpnex20";                        // The application to give focus to.

        private bool confirmRename = true;                                  // Whether the user should be prompted to confirm that the file is about to be renamed.
        private const bool confirmOverwrite = true;
        private bool useThumbnailMode = false;                              // Whether to use the thumbnail method to load the preview images.
        private bool autoRenameAllowedAutomated = true;                     // Whether to allow the the file to be automatically renamed if the file change was automated (Commandline Argument / SendMessage).
        private bool autoRenameAllowedManual = false;                       // Whether to allow the the file to be automatically renamed if the file was loaded by the user manually through the program.
        private bool autoRenameAllowedBatch = true;                         // Whether to allow the the file to be automatically renamed if the file was loaded in a batch.
        private const int AutoRenameSeconds = 15;                           // The number of seconds to elapse before the file is automatically renamed.
        private bool _autoRenameActive = false;                             // This indicates whether autoRename is turned on (I.E. the internal state of the autoRenamer and it should be the same at the autoRenameCheckBox.checked state). It is not whether the autoRenamer is counting down. Note: Do not modify directly instead use the AutoRenameActive property.
        private int _autoRenameCountDown = 0;                               // This should not be used except for in the AutoRenameCountDown property. Note: Do not modify directly instead use the AutoRenameCountDown property.

        ImageCache activeImage;                                             // The currently active image (I.E. The image that is displayed in the main image spot).

        Queue<BatchJob> batches = new Queue<BatchJob>();                       // Queue holding a list of batch jobs wating to be processed.

        private RenameProfile activeRenameProfile = null;                   // The currently active rename profile.
        private List<RenameProfile> renameProfiles = new List<RenameProfile>(); // All of the rename profiles.

        private enum MessageType : uint // The message type of the WM_COPYDATA message we are sending, mostly for future use.
        {
            UNSPECIFIED = 0,
            FILE_NAME,
            BATCH_FILE_NAMES
        }

        /* -- IMPORTANT --
         * -- IMPORTANT --
         * -- IMPORTANT --
         * 
         * This is a good convention that I should use always in future.
         * 
         * Note: The following propertys are used to perform actions when state variables are updated.
         * This is a new better way of managing things that need to be updated on the form state when 
         * state vairables change. This causes the change to directly update the form display state rather
         * then the caller being responcible for everything. Use this method entirely in future as it seems to work really well.
         * 
         * Convention basically works as follows: Interface generates a change in program state, change in state performs relevent actions and updates form display if necessary.
         * Rather then change in interface triggers actions and updates form state.
         * 
         * Note: Currently this program is half and half of the above two styles, this is because the superior way was only idetified part way through the 
         * implementation process. Any modifications should converge to the new method rather then the old.
         * 
         * -- IMPORTANT --
         * -- IMPORTANT --
         * -- IMPORTANT --
         */

        /// <summary>
        /// Gets or sets whether the autoRename is turned on (I.E. the internal state of the autoRenamer and it should be the same at the autoRenameCheckBox.checked state). It is not whether the autoRenamer is counting down.
        /// </summary>
        private bool AutoRenameActive
        {
            get
            {
                return _autoRenameActive;
            }
            set
            {
                _autoRenameActive = value;

                updateAutoRenameCaption();
            }
        }

        /// <summary>
        /// Gets or sets the current number of seconds that need to elapse befor the auto renamer will rename the file.
        /// </summary>
        private int AutoRenameCountDown
        {
            get
            {
                return _autoRenameCountDown;
            }
            set
            {
                _autoRenameCountDown = value;

                updateAutoRenameCaption();
            }
        }

        /// <summary>
        /// Construct the form.
        /// </summary>
        public MainForm()
        {
            InitializeComponent();
        }

        /// <summary>
        /// Called when the form is loaded.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void MainForm_Load(object sender, EventArgs e)
        {
            // Make the form occupy the right half of the screen.
            // Note: This may not be the right way to find the screen size.
            Rectangle screenSize = System.Windows.Forms.Screen.GetWorkingArea(this);
            this.Left = screenSize.Width / 2;
            this.Top = 0;
            this.Width = this.Left;
            this.Height = screenSize.Height;

            mainImage.AllowDrop = true; // A hidden property that allows the picture box to allow dragging and dropping.

            // Get the command line arguments.
            string[] cmd = Environment.GetCommandLineArgs();    // First argument will be the executable path/name and the seconds can be a file name to load.
            string messageData;
            List<string> filenames = new List<string>();
            MessageType externalType;

            // Check whether there are any files specified on the command line. (First argument is the exe path/name, subsequent arguments are assumed to be filenames).
            if (cmd.Length >= 2)
            {
                // Determine whether we are dealing with a single file or a batch of files.
                if (cmd.Length == 2)
                {
                    messageData = cmd[1];                   // Get the single file name from the second argument (the first is the exe path).
                    externalType = MessageType.FILE_NAME;   // Sending a single file.
                }
                else
                {
                    for (int i = 1; i < cmd.Length; i++)            // We need the list of filenames for two purposes, so we can mangle them into a single string or so we can batch process them.
                    {
                        filenames.Add(cmd[i]);
                    }

                    messageData = mangleBatchFilenames(filenames);  // Create a single string of all filenames.
                    externalType = MessageType.BATCH_FILE_NAMES;    // Sending a batch of files.
                }

                //if (sendStringMessage("ScanHelper", (uint)externalType, messageData))										// For use if testing out of the debugger.
                //if (sendStringMessage(Process.GetCurrentProcess().ProcessName.vshost, (uint)externalType, messageData))     // For use if testing in the with debugger.
                if (sendStringMessage(Process.GetCurrentProcess().ProcessName, (uint)externalType, messageData))
                {
                    // We have sent the file to an already running instance so close this instance.
                    Close();

                    return;
                }
                else
                {
                    if (externalType == MessageType.FILE_NAME)  // Use the external type "incorrectly" to determine whether we are single or batch processing, just easier and more accurate then any other method.
                    {
                        loadImage(messageData, autoRenameAllowedAutomated);
                    }
                    else
                    {
                        newBatch(filenames);
                    }
                }
            }

            // Initalise settings based on the original values (make sure the controls are displaying syncronised with there status variables).
            setConfirmRename(confirmRename);
            setUseThumbnailMode(useThumbnailMode);
            setAutoRename(AutoRenameActive);
            setAutoRenameAllowedAutomated(autoRenameAllowedAutomated);
            setAutoRenameAllowedManual(autoRenameAllowedManual);
            setFocusApplication(focusProcessName);

            // Load The Settings From File.
            loadSettings();
            allowSaveSettings = true;       // Allow this instance to save the settings back to file.

            // Make sure GUI particulars are up to date.
            updateBatchVisualState();
            redoSampleImageBoxes();

            // Give the rename button focus.
            renameButton.Select();
        }

        /// <summary>
        /// Called when the form has finnished resizing. We need to manually update the sample image boxes.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void MainForm_ResizeEnd(object sender, EventArgs e)
        {
            redoSampleImageBoxes();
        }

        /// <summary>
        /// The windows message loop function. We need to add to this function to intercept the WM_COPYDATA message used to transfer the filename between running instances.
        /// </summary>
        /// <param name="m"></param>
        protected override void WndProc(ref Message m)
        {
            //Note To Self: Don't use CopyData in managed code in future, a TCP connection would be simpler and nicer.
            if (m.Msg == WM_COPYDATA)   // If it is the copy data message.
            {
                CopyDataStruct data = (CopyDataStruct)m.GetLParam(typeof(CopyDataStruct));          // Get the copy data struct.

                processCopiedData(data);
            }

            base.WndProc(ref m);
        }

        /// <summary>
        /// Called when the form is closing. Used to save the settings to file when the application is closed.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void MainForm_FormClosing(object sender, FormClosingEventArgs e)
        {
            saveSettings();
        }

        /// <summary>
        /// Called when the MainImage is doubled clicked and is used to load a new file.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void mainImage_DoubleClick(object sender, EventArgs e)
        {
            OpenFileDialog ofd = new OpenFileDialog();

            ofd.CheckFileExists = true;
            ofd.CheckPathExists = true;
            ofd.Multiselect = false;
            ofd.ShowReadOnly = false;
            ofd.ValidateNames = true;

            DialogResult result = ofd.ShowDialog();

            if (result == DialogResult.OK)
            {
                loadImage(ofd.FileName, autoRenameAllowedManual);
            }
        }

        /// <summary>
        /// Called when the openFileButton is clicked and is used to load the main active image in the systems default application.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void openFileButton_Click(object sender, EventArgs e)
        {
            if (activeImage != null)
            {
                if (System.IO.File.Exists(activeImage.Filename))
                {
                    System.Diagnostics.Process.Start(activeImage.Filename);
                }
            }
        }

        /// <summary>
        /// Called when a preview image is clicked.
        /// 
        /// Note: The preview image name is used to identify the image from the cache it holds. The image from the cache is one less then the preview
        /// image name. This is not the most robust method of identifying the image, but it is otherwise difficult associate to two.
        /// </summary>
        /// <param name="sender">The picture that was clicked.</param>
        /// <param name="e"></param>
        public void PreviewImage_Click(object sender, EventArgs e)
        {
            PictureBox pictureBox = (PictureBox)sender;

            int index = 0;

            for (int i = 0; i < pictureBox.Name.Length; i++)
            {
                char nextChar = pictureBox.Name[i];
                if (char.IsDigit(nextChar))
                {
                    index = index * 10 + int.Parse(nextChar.ToString());
                }
            }

            changeActiveImage(index - 1);
        }

        /// <summary>
        /// Called when an object is draged over the mainImage.
        /// 
        /// Used to change the cursor so the user gets feedback if the right type of data is dragged.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void mainImage_DragEnter(object sender, DragEventArgs e)
        {
            // If the data is a file, display change the cursor.
            if (e.Data.GetDataPresent(DataFormats.FileDrop))
            {
                e.Effect = DragDropEffects.Link;
            }
            else
            {
                e.Effect = DragDropEffects.None;
            }

        }

        /// <summary>
        /// Called when the user drags and drops an object over the mainImage.
        /// 
        /// Used to grab and open the file if the object has a filename.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void mainImage_DragDrop(object sender, DragEventArgs e)
        {
            // If the data is or can be converted to a filename.
            if (e.Data.GetDataPresent("FileName"))
            {
                object dropData = e.Data.GetData("FileName");   // Get a list of filenames (note only ever seems to be one but we handel the case for many anyway).
                if (dropData is string[])                       // Type check the data.
                {
                    string[] fileNames = (string[])dropData;    // Convert to array of strings type.

                    for (int i = 0; i < fileNames.Length; i++)  // For all of the filename strings.
                    {
                        loadImage(fileNames[i], autoRenameAllowedManual);   // Load the files.
                    }
                }
            }
        }

        /// <summary>
        /// Called when the load batch button is clicked. Processes the next batch job.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void loadBatchButton_Click(object sender, EventArgs e)
        {
            loadBatch();
        }

        /// <summary>
        /// Called when the previous preview button is clicked.
        /// 
        /// Note: This is currnetly not implemented. But should be implemented to go to the next previous image in the current directory.
        /// Implementation Ideas: If the last preview box is empty fill it, if it has an image cycle the image and put it in the free spot created.
        /// Search through the directory for the last file from the cache and then load the next one, wrapping around if it is at the end.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void PreviousPreviewButton_Click(object sender, EventArgs e)
        {
            // TODO
            MessageBox.Show("Not Implemented.");
        }

        /// <summary>
        /// Called when the next preview button is clicked.
        /// 
        /// Note: This is currnetly not implemented. But should be implemented to go to the next previous image in the current directory.
        /// Implementation Ideas: If the last preview box is empty fill it, if it has an image cycle the image and put it in the free spot created.
        /// Search through the directory for the last file from the cache and then load the next one, wrapping around if it is at the end.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void NextPreviewButton_Click(object sender, EventArgs e)
        {
            // TODO
            MessageBox.Show("Not Implemented.");
        }

        /// <summary>
        /// Called when the info status label is clicked and is used to manually change the info status displayed (also disables the auto update if it is running).
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void statusLabel_Click(object sender, EventArgs e)
        {
            if (autoUpdateInfoStatus)
            {
                autoUpdateInfoStatus = false;
            }
            else
            {
                nextInfoStatus();
            }
        }

        /// <summary>
        /// Called when the info status label is double clicked and is used to toggle the auto update of the info status label.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void statusLabel_DoubleClick(object sender, EventArgs e)
        {
            autoUpdateInfoStatus = !autoUpdateInfoStatus;
        }

        /// <summary>
        /// Opens the directory activeImages is in using explorer. 
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void directoryButton_Click_1(object sender, EventArgs e)
        {
            if (activeImage != null)
            {
                string directory = System.IO.Path.GetDirectoryName(activeImage.Filename);

                if (System.IO.Directory.Exists(directory))
                {
                    System.Diagnostics.Process.Start("explorer", directory);
                }
            }
        }

        /// <summary>
        /// Called when the delete button is clicked. This deletes the main image.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void DeleteButton_Click(object sender, EventArgs e)
        {
            deleteActiveImage();
        }

        /// <summary>
        /// Called when the AutoRenameTimer ticks over. Used to count down the seconds and the auto rename the file when the count reaches zero.
        /// 
        /// Note: The timer should tick at 1 second intervals.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void AutoRenameTimer_Tick(object sender, EventArgs e)
        {
            if (AutoRenameActive == true)
            {
                if (!pauseCheckBox.Checked)
                {
                    AutoRenameCountDown--;
                }

                if (AutoRenameCountDown == 0)
                {
                    // Stop the auto renamer first. 
                    // We do it first incase the rename operation takes some time and the auto renamer kicks in befor we have a chance to stop it.
                    startStopAutoRename(false);

                    autoRename();
                }
            }
            else
            {
                #if DEBUG
                if (System.Diagnostics.Debugger.IsAttached)
                {
                    MessageBox.Show("Warning: AutoRenameTimer enabled but autoRenameActive is false.");
                }
                #endif

                // This is a safe guard, it should never actually have to be called, and is almost certainly a logic error if it is.
                startStopAutoRename(false);
            }
        }

        /// <summary>
        /// Called when the auto update info status timer expires and it is time to move to the next status.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void statusTimer_Tick(object sender, EventArgs e)
        {
            if (autoUpdateInfoStatus)
            {
                nextInfoStatus();
            }
        }

        /// <summary>
        /// Called when the autoRename checkbox is clicked.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void AutoRenameCheckBox_CheckedChanged(object sender, EventArgs e)
        {
            setAutoRename(autoRenameCheckBox.Checked);
        }

        /// <summary>
        /// Called when the rename button is clicked.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void RenameButton_Click(object sender, EventArgs e)
        {
            rename();
        }

        /// <summary>
        /// Called when the renameConfirmCheckBox is clicked.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void RenameConfirmCheckBox_CheckedChanged(object sender, EventArgs e)
        {
            setConfirmRename(renameConfirmCheckBox.Checked);
        }

        /// <summary>
        /// Called when the thumbnailModeCheckBox is clicked.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void thumbnailModeCheckBox_CheckedChanged(object sender, EventArgs e)
        {
            setUseThumbnailMode(thumbnailModeCheckBox.Checked);
        }

        /// <summary>
        /// Called when the autoRenameCmdCheckBox is changed. Is used to change the internal state.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void autoRenameCmdCheckBox_CheckedChanged(object sender, EventArgs e)
        {
            setAutoRenameAllowedAutomated(autoRenameCmdCheckBox.Checked);
        }

        /// <summary>
        /// Called when the autoRenameManualCheckBox is changed. Is used to change the internal state
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void autoRenameManualCheckBox_CheckedChanged(object sender, EventArgs e)
        {
            setAutoRenameAllowedManual(autoRenameManualCheckBox.Checked);
        }

        /// <summary>
        /// Called when the autoRenameCmdCheckBox is actually clicked (and not changed by other means) and is used to warn the user how the item works.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void autoRenameCmdCheckBox_Click(object sender, EventArgs e)
        {
            warnAutoRenameChangedAllowed();
        }

        /// <summary>
        /// Called when the autoRenameManualCheckBox is actually clicked (and not changed by other means) and is used to warn the user how the item works.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void autoRenameManualCheckBox_Click(object sender, EventArgs e)
        {
            warnAutoRenameChangedAllowed();
        }

        /// <summary>
        /// Called when the focus process name is changed.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void focusProcessTextBox_TextChanged(object sender, EventArgs e)
        {
            setFocusApplication(focusProcessTextBox.Text);
        }

        /// <summary>
        /// Called when the BaseDirectoryTextBox text changes and is used to update the base directory.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void BaseDirectoryTextBox_TextChanged(object sender, EventArgs e)
        {
            RenameProfile.CommonDirectory = baseDirectoryTextBox.Text;
        }

        /// <summary>
        /// Called when the BaseDirectoryButton (...) button is clicked and is used to display the folder selection dialog.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void BaseDirectoryButton_Click(object sender, EventArgs e)
        {
            folderBrowserDialog.SelectedPath = baseDirectoryTextBox.Text;
            folderBrowserDialog.ShowDialog();
            baseDirectoryTextBox.Text = folderBrowserDialog.SelectedPath;

            simplifyPath();  // Note: By design the path is only simplified when the directory picker is used, it is assumed that if the user enters it they know best and as such may not want to simplify it.
        }

        /// <summary>
        /// Called when the DirectoryButton (...) button is clicked and is used to display the folder selection dialog.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void DirectoryButton_Click(object sender, EventArgs e)
        {
            folderBrowserDialog.SelectedPath = specificDirectoryTextBox.Text;
            folderBrowserDialog.ShowDialog();
            specificDirectoryTextBox.Text = folderBrowserDialog.SelectedPath;

            simplifyPath(); // Note: By design the path is only simplified when the directory picker is used, it is assumed that if the user enters it they know best and as such may not want to simplify it.
        }

        /// <summary>
        /// Called when the category is changed.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void CategoryComboBox_SelectedIndexChanged(object sender, EventArgs e)
        {
            if ((activeRenameProfile != null) && (activeRenameProfile.Category != categoryComboBox.Text))   // If the active rename profile does not belong to the newly selected category.
            {
                activeRenameProfile = null;     // Remove the active rename profile, so that the first rename profile that belongs to the category is selected later on.
            }

            refreshProfilesComboBox();

            if ((activeRenameProfile == null) || (activeRenameProfile.Category != categoryComboBox.Text))
            {
                profileNameComboBox.SelectedIndex = 0;
            }
        }

        /// <summary>
        /// Called when the current rename profile is changed.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void ProfileNameComboBox_SelectedIndexChanged(object sender, EventArgs e)
        {
            updateActiveRenameProfile();
        }

        /// <summary>
        /// Called when the AutoUpdateCheckBox changes.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void AutoUpdateCheckBox_CheckedChanged(object sender, EventArgs e)
        {
            if (activeRenameProfile != null)
            {
                activeRenameProfile.AutoUpdateSequenceMask = autoUpdateCheckBox.Checked;
            }
        }

        /// <summary>
        /// Called when the ResetButton is clicked.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void ResetButton_Click(object sender, EventArgs e)
        {
            if (activeRenameProfile != null)
            {
                changeActiveRenameProfileSettings();    // Make sure the active profile is up to date.
                activeRenameProfile.SequenceNumber = 1; // Set the sequence number.
                refreshActiveRenameProfile();           // Reload the new mask into the update controls.
                changeActiveRenameProfileSettings();    // Force the settings to be updated with the new number.
            }
        }

        /// <summary>
        /// Clicked when the AddButton is clicked to add a new rename profile.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void AddButton_Click(object sender, EventArgs e)
        {
            createNewRenameProfile();
        }

        /// <summary>
        /// Called when the UpdateButton is clicked to update the current rename profile.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void UpdateButton_Click(object sender, EventArgs e)
        {
            changeActiveRenameProfileSettings();
        }
        
        /// <summary>
        /// Called when the RemoveButton is clicked to remove the active rename profile.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void RemoveButton_Click(object sender, EventArgs e)
        {
            removeActiveRenameProfile();
        }

        /// <summary>
        /// Loads a new image as the main image and puts it into the image cache.
        /// </summary>
        /// <param name="imageFileName">The file name of the image to load.</param>
        /// <param name="allowRename">Whether this new image is allowed to be renamed.</param>
        /// <returns>Whether the image was able to be loaded.</returns>
        private bool loadImage(string imageFileName, bool allowRename)
        {
            ImageCache loadedImageData = new ImageCache();

            autoRename();  // Auto rename the last image if it supposed to be renamed.

            // Try To Load The New Image
            try
            {
                Image handleFreeImage;

                if (!useThumbnailMode)
                {
                    // Because of this two stage load file->memory->image the application uses double the images original memory.
                    // But this seems to be the only way since Image.From* seems to force a retained handle from what it was created from but still copies the data.
                    // The first case locks access to the file (which we can't have) and the second locks the memory stream (which leads to wasted memory).
                    System.IO.MemoryStream ms = new System.IO.MemoryStream(System.IO.File.ReadAllBytes(imageFileName));
                    handleFreeImage = Image.FromStream(ms);
                }
                else
                {
                    // This method generates a small thumbnail image copy, the problem is if the image has a true thumbnail that will be used and can often be substandard for this purpose.
                    // We use mainImage dimensions as it is the biggest image on the form. Can lead to less then standard images if the form is resized, but is 
                    // simpler and cost efficient then reloading them and only effects already loaded images, could make it reload if resized at a later date.
                    Image thumbnail = Image.FromFile(imageFileName);
                    float newWidth = mainImage.Width;
                    float newHeight = mainImage.Height;
                    aspectCorrectResize(thumbnail.Width, thumbnail.Height, ref newWidth, ref newHeight);
                    handleFreeImage = thumbnail.GetThumbnailImage((int)newWidth, (int)newHeight, null, IntPtr.Zero);
                    thumbnail.Dispose();    // Explicitly dispose of the image so that it can be garbage collected asap.
                }
                
                loadedImageData.Image = handleFreeImage;
                loadedImageData.Filename = imageFileName;
                loadedImageData.AutoRenameable = allowRename;
                fileLoadedStats();
                warnImageFormateChanged(System.IO.Path.GetExtension(imageFileName));
            }
            catch (Exception e)
            {
                MessageBox.Show("Error loading image. Error: " + e.Message);

                return false;
            }

            // If the cache is full, kill the last image in the cache.
            if (imageCache[(imageCacheSize - 1)] != null)
            {
                if (imageCache[(imageCacheSize - 1)].Image != null)
                {
                    // Explicitly dispose of the image so that it can be garbage collected asap.
                    imageCache[(imageCacheSize - 1)].Image.Dispose();
                }
            }

            // Move all the images down in the cache.
            for (int i = 0; i < imageCacheSize - 1; i++)
            {
                imageCache[(imageCacheSize - 1) - i] = imageCache[(imageCacheSize - 1) - i - 1];
            }
            imageCache[0] = loadedImageData;        // Add the loaded image to the start of the cache.
            
            changeActiveImage(0);                   // Changes the active image to the image that was just loaded and put in the cache.

            if (allowRename)
            {
                startStopAutoRename(true);          // Start the auto rename process for the new image if it allows it.
            }

            // Give The Scan Driver Focus.
            Process[] process = System.Diagnostics.Process.GetProcessesByName(focusProcessName);
            if (process.Length > 0)
            {
                SetForegroundWindow(process[0].MainWindowHandle);
            }

            // Force garbage collection to free memory from unused images.
            GC.Collect();

            return true;
        }
        
        /// <summary>
        /// Corrects the aspect ratio of the out parameters using the in parameters. This can be usefull when resizing images.
        /// </summary>
        /// <param name="widthIn">The original width of the image.</param>
        /// <param name="heightIn">The original height of the image.</param>
        /// <param name="widthOut">Passed in the container size, and out is the aspect corrected size.</param>
        /// <param name="heightOut">Passed in the container size, and out is the aspect corrected size.</param>
        void aspectCorrectResize(float widthIn, float heightIn, ref float widthOut, ref float heightOut)
        {
            float aspectFit = widthOut / heightOut;
            float aspectIn = widthIn / heightIn;

            if (aspectIn <= aspectFit)
            {
                //heightOut = heightOut;
                widthOut = heightOut * aspectIn;
            }
            else
            {
                //widthOut = widthOut;
                heightOut = (float)(widthOut * (1.0 / aspectIn));
            }
        }

        /// <summary>
        /// Changes the active image to one of the other image in the image cache.
        /// </summary>
        /// <param name="index">The index of the image in the cache (0 indexded).</param>
        private void changeActiveImage(int index)
        {
            if ((index >= 0) && (index < imageCacheSize))
            {
                activeImage = imageCache[index];
                updateImages();
            }
        }

        /// <summary>
        /// Update all images on the form using the images in the cache as previews and the activeImage as the main image.
        /// </summary>
        private void updateImages()
        {
            Image updateImage = null;               // This is used to update an image, null is used to clear the image if one has already been loaded, but none should be loaded.

            updateInfoLabels();                     // If we are changing the pictures we also need to update the info labels.

            updateImage = null;                     // If we don't load an image we want to clear it.
            if (activeImage != null)                // But...If have an activeImage.
            {
                updateImage  = activeImage.Image;   // We want to set the image to the active image. 
            }
            setImage(mainImage, updateImage);       // Set the mainImage.

            // Update the displayed images in the previews from the images in the image cache.
            for (int i = 0; i < previewImages(); i++)
            {
                PictureBox p = previewImage(i);
                if (p != null)
                {
                    updateImage = null;             // If we don't load an image we want to clear it.
                    if (imageCache[i] != null)      // But...If have an image at this spot in the cache.
                    {
                        updateImage = imageCache[i].Image;  // We want to set the image to the one from the image cache.
                    }
                    setImage(p, updateImage);       // Set this preview image
                }
                else
                {
                    // We have a problem if this ever happens because we should be preventing this, and is a logic error.
#if DEBUG
                    if (System.Diagnostics.Debugger.IsAttached)
                    {
                        MessageBox.Show("Invalid preview image attempted.");
                    }
#endif
                }
            }
        }

        /// <summary>
        /// Loads the specified Image into the specified PictureBox.
        /// 
        /// To clear the picture box set image to null.
        /// </summary>
        /// <param name="pb">The PictureBox to update.</param>
        /// <param name="img">The image to load.</param>
        private void setImage(PictureBox pb, Image img)
        {
            if (img != null)
            {
                // No longer used but is still a good routine for determining an aspect maintained size that will fit into a container.
                /*
                    float aspectPB = (float)pb.Width / pb.Height;
                    float aspectIMG = (float)img.Width / img.Height;

                    int width = pb.Width;
                    int height = pb.Height;

                    if (aspectIMG <= aspectPB)
                    {
                        height = pb.Height;
                        width = (int)(pb.Height * aspectIMG);
                    }
                    else
                    {
                        width = pb.Width;
                        height = (int)(pb.Width * (1.0 / aspectIMG));
                    }
                */

                pb.SizeMode = PictureBoxSizeMode.Zoom;
                pb.Image = img;
            }
            else
            {
                pb.Image = null;
            }
        }

        /// <summary>
        /// Gets the number of preview images on the form.
        /// </summary>
        /// <returns>The number of preview images.</returns>
        private int previewImages()
        {
            return previewPanel.Controls.Count;
        }

        /// <summary>
        /// Gets the picture box associated with the specified index (I.E. Makes the preview images index addressable.
        /// 
        /// Use in conjunction with previewImages().
        /// </summary>
        /// <param name="index">The index of the preview image picture box (starting at 0).</param>
        /// <returns>The picture box.</returns>
        private PictureBox previewImage(int index)
        {
            if (index < previewImages())
            {
                return (PictureBox)previewPanel.Controls[index]; // Could be a problem if we add something to this control that isn't a PictureBox.
            }

            return null;
        }

        /// <summary>
        /// Regenerates the preview image boxes.
        /// </summary>
        private void redoSampleImageBoxes()
        {
            const int previewSpacing = 10;              // The horizontal space between the boxes in pixels.
            const int previewWidth = 100;               // The width of the preview images.
            int previewHeight = previewPanel.Height;    // The height of the preview images.

            previewPanel.Controls.Clear(); // Could just add or remove, but since we are going to reload all images there is little point.

            int previews = (previewPanel.Width - previewWidth) / (previewWidth + previewSpacing) + 1;           // Calculate how many previews we can fit. The funny look of this calculation is because of picket fence picket = picture box, and rails = gap.

            if (previews > imageCacheSize)              // If we can display more images then we have in the cache clamp to the size of the cache.
            {
                previews = imageCacheSize;
            }

            int endSpacing = previewPanel.Width - (previewWidth * previews) - previewSpacing * (previews - 1);  // Calculate the space at each end of the preview image set in pixels.

            for (int i = 0; i < previews; i++)          // For all of the preview images.
            {
                PictureBox p = new PictureBox();                                                                    // Create a new picture box.
                p.Location = new System.Drawing.Point(endSpacing / 2 + i * (previewWidth + previewSpacing), 0);     // Put it in the right spot.
                p.Size = new System.Drawing.Size(previewWidth, previewHeight);                                      // Make it the right size.
                p.BorderStyle = System.Windows.Forms.BorderStyle.FixedSingle;                                       // Set the border style.
                p.Name = "previewImage" + (i + 1);                                                                  // Name it according to its index. Note: IMPORTANT This is used to identify which picture should become the active image when the picture box is clicked. See the Clicked delegate.
                p.TabStop = false;                                                                                  // Don't let it be a tab stop.
                p.Click += new System.EventHandler(PreviewImage_Click);                                             // Add a clicked event handeler (delegate).
                previewPanel.Controls.Add(p);                                                                       // Add it to the preview panel region.
            }

            updateImages();     // Load all of the preview boxes with images from the cache.
        }

        /// <summary>
        /// Deletes the active image.
        /// </summary>
        private void deleteActiveImage()
        {

            if (activeImage != null)
            {
                // Delete the images file.
                if (!deleteFile(activeImage.Filename))
                {
                        MessageBox.Show("Unable to delete image.");
                        return;
                }

                // Remove the active image from the cache.
                bool moveDown = false;
                for (int i = 0; i < imageCacheSize; i++)
                {
                    if ((imageCache[i] == activeImage) && (imageCache[i] != null))  // If we have found the image in the cache.
                    {
                        moveDown = true;                                            // Flag so that we can move all the images down in the cache.

                        if (imageCache[i].Image != null)
                        {
                            imageCache[i].Image.Dispose();                          // Explicitly dispose of the image so that it can be garbage collected asap.
                        }
                    }
                    if ((moveDown == true) && (i < (imageCacheSize - 1)))           // If we have found the image and are moving all the images down and we haven't reached the last image in the cache (which we can't move down).
                    {
                        imageCache[i] = imageCache[i + 1];                          // Move it down one spot.
                    }
                }

                // If the image was found in the cache the last spot will empty.
                if (moveDown)
                {
                    imageCache[imageCacheSize - 1] = null;
                }
            }

            activeImage = imageCache[0];    // Make the active image the bottom of the cache (if there are none in the cache it will just make it null).
            updateImages();                 // Update all the images on the form to reflect the changes.
        }

        /// <summary>
        /// Deletes a file.
        /// 
        /// Possible enhancement: Make this function so that it just deletes files to the recycling bin. Fairly easy just out of time, use win32 API not VB method.
        /// </summary>
        /// <param name="path">The file to delete.</param>
        /// <returns>Whether the file was successfully deleted.</returns>
        private bool deleteFile(string path)
        {
            bool success = false;

            if (System.IO.File.Exists(path))
            {
                try
                {
                    /* Previous Implementation - Before I could recycle files. Left incase I want to add other options in the future (reuse this code).
                    DialogResult result = MessageBox.Show("Currently I can't send the file to the recycling bin. \r\n" + 
                                        "Would you like to move the file to a dump folder on the desktop or delete the file permanently?\r\n\r\n" + 
                                        "Yes = Delete permanently!\r\n" + 
                                        "No = Move to dump folder on the desktop.\r\n" + 
                                        "Cancel = What I didn't mean to delete a file? Just leave it.",  
                                        "Delete File", MessageBoxButtons.YesNoCancel);
                    switch (result)
                    {
                        case DialogResult.Yes:
                            System.IO.File.Delete(path);
                            success = true;
                            break;
                        case DialogResult.No:
                            System.IO.Directory.CreateDirectory(dumpFolder);                                                    // Create the dump folder if it doesn't exist.
                            System.IO.File.Move(path, System.IO.Path.Combine(dumpFolder, System.IO.Path.GetFileName(path)));    // Move the file to the dump folder.
                            success = true;
                            break;
                        case DialogResult.Cancel:
                            break;
                    }
                    */

                    System.IO.FileInfo fsi = new System.IO.FileInfo(path);
                    success = ShellInteraction.FileFolderShellInteraction.recycle(fsi, this.Handle, true);

                    if (success)
                    { 
                        // Check the file is actually gone, the user may have selected not allowed the delete.
                        if (System.IO.File.Exists(path))
                        {
                            success = false;
                        }
                    }
                }
                catch (Exception e)
                {
                    // Failed deleting the file.
                }
            }

            return success;
        }

        /// <summary>
        /// Warns the user if the format of the file has changed (useful in detecting human error).
        /// </summary>
        /// <param name="newFormat">The format of the file that has just been loaded.</param>
        private void warnImageFormateChanged(string newFormat)
        {
            if (WarnImageTypeChanged)
            {
                if (lastImageFormatValid)
                {
                    if (newFormat != lastImageFormat)
                    {
                        MessageBox.Show("Warning! The loaded image differs in format from the previous image.");
                    }
                }

                lastImageFormat = newFormat;
                lastImageFormatValid = true;
            }
        }

        /// <summary>
        /// Use this function to indicate that a new file has just been loaded. It takes care of updating the status counters.
        /// 
        /// This function manages a global count. It also manages a queue of minute long bins which contain how many files were processed in that minute.
        /// After a minute has elapsed the queue is cycled, droping the oldest minute and adding a fresh blank one.
        /// </summary>
        private void fileLoadedStats()
        {
            long currentMinute = (long)(DateTime.Now.Subtract(startTime).TotalMinutes);

            if (statsLastMinute != currentMinute)
            {
                statsLastMinute = currentMinute;
                for (int i = (ShortAverageDuration - 1); i > 0; i--)
                {
                    filesOpenedMinute[i] = filesOpenedMinute[i - 1];
                }
                filesOpenedMinute[0] = 0;
            }

            filesOpened++;
            filesOpenedMinute[0]++;
        }

        /// <summary>
        /// Cycles to the next info status to be displayed and updates it on the form.
        /// </summary>
        private void nextInfoStatus()
        {
            currentInfoStatus = (currentInfoStatus + 1) % TotalInfoStatus;
            updateInfoStatus();
        }

        /// <summary>
        /// Updates the info status label based on the the currentInfoStatus.
        /// </summary>
        private void updateInfoStatus()
        {
            if (activeImage != null)
            {
                string statusString = "";
                switch (currentInfoStatus)
                {
                    case 0:
                        string space = @"N\A";

                        if (System.IO.Path.IsPathRooted(activeImage.Filename))
                        {
                            long freeBytes = new System.IO.DriveInfo(System.IO.Path.GetPathRoot(activeImage.Filename)).TotalFreeSpace;

                            float usefulBytes = freeBytes / (1024 * 1024); // Convert to MB
                            if (usefulBytes >= 1024)
                            {
                                space = Math.Round(usefulBytes / 1024, 3) + " GB";
                            }
                            else
                            {
                                space = Math.Round(usefulBytes, 3) + " MB";
                            }
                        }

                        statusString = "Free Drive Space: " + space;
                        break;
                    case 1: statusString = "Total Files Loaded: " + filesOpened;
                        break;
                    case 2: statusString = "Files Per Hour (Total): " + Math.Round(filesOpened / ((float)(DateTime.Now.Subtract(startTime).TotalSeconds) / 3600));
                        break;
                    case 3:
                        int total = 0;

                        for (int i = 0; i < ShortAverageDuration; i++)
                        {
                            total += filesOpenedMinute[i];
                        }

                        statusString = "File Per Hour (Last 10 Minutes): " + Math.Round((float)total * (60.0 / ShortAverageDuration));
                        break;
                    // Note: If adding more info status need to increase TotalInfoStatus.
                }

                statusLabel.Text = statusString;
            }
        }

        /// <summary>
        /// Updates the auto rename caption.
        /// </summary>
        private void updateAutoRenameCaption()
        {
            const string BaseCaption = "Auto Rename";

            //if (AutoRenameTimer.Enabled)      // Only Diplay the count down when the timer is active. (The choice between the two is really personal preference.)
            if (AutoRenameActive)               // Display the count down when active regardless of whether it is counting down or not (will just hold at the start value).
            {
                autoRenameCheckBox.Text = BaseCaption + " (" + AutoRenameCountDown.ToString() + ")";
            }
            else
            {
                autoRenameCheckBox.Text = BaseCaption;
            }
        }

        /// <summary>
        /// Sets the auto renamer so that it is turned on.
        /// 
        /// Note: There is a difference between the autorenamer being active and started. Started means that it is counting down, active means that it is in the state 
        /// where it will count down the next time an image triggers it. When seting the auto renamer on the renamer is also started (if possible) so it can begin 
        /// counting down to rename the current active image.
        /// </summary>
        /// <param name="on">Whether to activate and start the auto renamer (true) or stop it (false).</param>
        private void setAutoRename(bool on)
        {
            if ((on == true) && (activeRenameProfile == null))
            {
                MessageBox.Show("Can't select auto mode with no active rename profile.");
                on = false;
            }

            AutoRenameActive = on;
            startStopAutoRename(on);
            autoRenameCheckBox.Checked = on;

        }

        /// <summary>
        /// Resets the auto renamer (speifically its timed countdown).
        /// </summary>
        private void resetAutoRename()
        {
            AutoRenameCountDown = AutoRenameSeconds;
            autoRenameTimer.Interval = 1000;            // Tigger interupts at 1 second intervals.
        }

        /// <summary>
        /// Starts or stops the auto rename countdown.
        /// 
        /// If autoRenamer isn't active or the active image is not renameable the counter will not start irrespective of the state See: setAutoRename().
        /// 
        /// Start starts the timer counting down. If it is already running it does not reset it.
        /// Note: This just effects the count down and does not change the state of the auto renamer (active or not).
        /// </summary>
        /// <param name="start">Whether to start the count down (true) or stop it (false).</param>
        private void startStopAutoRename(bool start)
        {
            if ((AutoRenameActive == true) && (activeImage != null) && (activeImage.AutoRenameable == true) && (start == true))
            {
                if (autoRenameTimer.Enabled == false)   // If we arn't counting down we don't want to restart the timer if we already are.
                {
                    resetAutoRename();                  // Reset the timer.
                }

                autoRenameTimer.Enabled = true;         // Start the count down.
            }
            else
            {
                autoRenameTimer.Enabled = false;        // Stop the count down.
                resetAutoRename();                      // Reset the counter back to the start (to leave it in a state ready to go and as a safe guard).
            }
        }

        /// <summary>
        /// The auto renamer has activated that the current active image should be renamed. This renames the active image and stops the count down.
        /// </summary>
        private void autoRename()
        {
            if (AutoRenameActive)
            {
                // Stop the auto renamer first. Do NOT change this. 
                // If we have auto renamed the file we don't want the auto renamer still running in case it fires multiple times. 
                // We do it first incase the rename operation takes some time and the auto renamer kicks in befor we have a chance to stop it.
                startStopAutoRename(false);

                if (activeImage != null)
                {
                    if (activeImage.AutoRenameable)
                    {
                        rename();
                    }
                }
            }
        }

        /// <summary>
        /// Renames the active image using the active rename profile.
        /// </summary>
        /// <returns>Whether the file was sucessfully renamed.</returns>
        private bool rename()
        {
            if ((activeRenameProfile != null) && (activeImage != null))
            {
                try
                {
                    // Stop the auto renamer first. It may be running it may not be, but better to be safe then sorry and no harm stopping it if its not running.
                    // Don't want the auto renamer to attempt to rename the file if we have manually renamed it.
                    // We do it first incase the rename operation takes some time and the auto renamer kicks in befor we have a chance to stop it.
                    startStopAutoRename(false);

                    string oldFileName = activeImage.Filename;                                                              // Get the images current name.
                    string newFileName = activeRenameProfile.getNextFileName(System.IO.Path.GetExtension(oldFileName));     // Generate a new name for the image.
                    
                    if (newFileName == null) // If we failed to generate a valid new file name.
                    {
                        MessageBox.Show("Failed to generate a vaild new file name. Please check the parameters and try again.");
                        return false;
                    }

                    if (confirmOverwrite)   // If we want to confirm if the user is overwriting the file, check and warn them if we will.
                    {
                        if (System.IO.File.Exists(newFileName))
                        {
                            DialogResult result = MessageBox.Show("The file you are renameing to, if you continue you will overwrite the following file: \r\n" + newFileName + "\r\n\r\n Do you want to overwrite the file?", "Confirm Overwrite", MessageBoxButtons.YesNo);
                            if (result == DialogResult.No)
                            {
                                return false;
                            }
                            else
                            {
                                if (!deleteFile(newFileName))
                                {
                                    MessageBox.Show("Unable overwrite the file.");
                                    return false;
                                }
                            }
                        }
                    }

                    if (confirmRename)      // If we want to confirm if before we rename, check and only proceed if we get the ok.)
                    {
                        DialogResult result = MessageBox.Show("About to rename:\r\n" + oldFileName + "\r\nTo:\r\n" + newFileName + "\r\n\r\n Do you want to rename the file?", "Confirm Rename", MessageBoxButtons.YesNo);
                        if (result == DialogResult.No)
                        {
                            return false;
                        }
                    }

                    if (System.IO.File.Exists(oldFileName))                                             // Sanity check that the file still exists and something hasn't happened to it.
                    {
                        string newDirectory = System.IO.Path.GetDirectoryName(newFileName);             // Extract the directory for the new image.

                        if (newDirectory != "")                                                         // If the directory isn't blank.
                        {
                            System.IO.Directory.CreateDirectory(newDirectory);                          // Create the directory if it doesn't exist.
                        }
                        if (actuallyRename)                                                                 // This is a debugging mechanism used to turn renaming the file on or off depending on whether we want to acutally rename the file (if every thing is working, or not if we are testing problems elsewhere).
                        {
                            System.IO.File.Move(oldFileName, newFileName);                              // Move the file...the whole point of the exercise.
                        }
                        activeImage.Filename = newFileName;     // Update the caches file name (not because active image is a reference it updates the only (cache...whatever) copy).
                        activeImage.AutoRenameable = false;     // The image has been renamed once, we don't want to allow it to be auto renamed again (can be manually renamed).
                        activeRenameProfile.filenameUsed();     // Indicate that we have used the generated filename.
                        refreshActiveRenameProfile();           // Update the rename profile incase the sequence number has changed.
                    }
                    else
                    {
                        MessageBox.Show("Can't find the file to rename.");
                    }
                }
                catch (Exception e)
                {
                    MessageBox.Show("Error renaming file. Error: " + e.Message);
                }
            }

            return true;
        }

        /// <summary>
        /// Sets the confirmRename variable. Note all changes of this variable should be done through this function.
        /// 
        /// Note: This function also takes care of the gui update.
        /// </summary>
        /// <param name="confirm">The value to set confirmRename to.</param>
        private void setConfirmRename(bool confirm)
        {
            confirmRename = confirm;
            renameConfirmCheckBox.Checked = confirm;
        }

        /// <summary>
        /// Sets the useThumbnailMode variable. Note all changes of this variable should be done through this function.
        /// 
        /// Note: This function also takes care of the gui update.
        /// </summary>
        /// <param name="confirm">The value to set useThumbnailMode to.</param>
        private void setUseThumbnailMode(bool thumbnailMode)
        {
            useThumbnailMode = thumbnailMode;
            thumbnailModeCheckBox.Checked = thumbnailMode;
        }

        /// <summary>
        /// Sets the autoRenameAllowedAutomated variable. Note all changes of this variable should be done through this function.
        /// 
        /// Note: This function also takes care of the gui update.
        /// </summary>
        /// <param name="state">The value to set autoRenameAllowedAutomated to.</param>
        private void setAutoRenameAllowedAutomated(bool state)
        {
            // Note: This will only effect files loaded from now on.
            autoRenameAllowedAutomated = state;
            autoRenameCmdCheckBox.Checked = state;
        }

        /// <summary>
        /// Sets the setAutoRenameAllowedManual variable. Note all changes of this variable should be done through this function.
        /// 
        /// Note: This function also takes care of the gui update.
        /// </summary>
        /// <param name="state">The value to set setAutoRenameAllowedManual to.</param>
        private void setAutoRenameAllowedManual(bool state)
        {
            // Note: This will only effect files loaded from now on.
            autoRenameAllowedManual = state;
            autoRenameManualCheckBox.Checked = state;
        }

        /// <summary>
        /// Sets the focusProcessName variable. Note all changes of this variable should be done through this function.
        /// 
        /// Note: This function also takes care of the gui update.
        /// </summary>
        /// <param name="processName">The value to set focusProcessName to.</param>
        private void setFocusApplication(string processName)
        {
            focusProcessName = processName;

            focusProcessTextBox.Text = focusProcessName;
        }

        /// <summary>
        /// Warns the user how the auto renamer options work.
        /// </summary>
        private void warnAutoRenameChangedAllowed()
        {
            string message = "This will only effect files loaded from now on.";

            if (autoRenameCheckBox.Checked == false)
            {
                message += "\r\n\r\n" + "Note you must still turn on the auto renamer.";
            }

            MessageBox.Show(message);
        }

        /// <summary>
        /// Updates the GUI info labels.
        /// </summary>
        private void updateInfoLabels()
        {
            if (activeImage != null)
            {
                mainImageLable.Text = activeImage.Filename + @" (Length: " + activeImage.Filename.Length + @")";
            }
            else
            {
                mainImageLable.Text = "";
            }


            if ((activeRenameProfile != null) && (activeImage != null))
            {
                renameInfoLabel.Text = "Renaming: " + activeImage.Filename + "\r\n\r\nTo: " + activeRenameProfile.getNextFileName(System.IO.Path.GetExtension(activeImage.Filename));
            }
            else
            {
                renameInfoLabel.Text = "";
            }
        }

        /// <summary>
        /// Updates the CategoryComboBox if it is not already set. This saves a lot of unnecessary calls generated by the GUI switching to the same thing.
        /// 
        /// This function could be replaced with a function that took no argument and just used the activeRenameProfile as all calls currently 
        /// are based on this, but this allows for future expansion. If needed this function can be adapted in this way.
        /// This call mirrors smartSetProfileNameComboBox() which is used to resolve a bug.
        /// </summary>
        /// <param name="text">The text to set.</param>
        private void smartSetCategoryComboBox(string text)
        {
            //if (categoryComboBox.Text != text)
            {
                categoryComboBox.Text = text;
            }
        }

        /// <summary>
        /// Updates the CategoryComboBox if it is not already set. This saves a few problems and also saves a lot of unnecessary calls generated by the GUI switching to the same thing.
        /// 
        /// This function could be replaced with a function that took no argument and just used the activeRenameProfile as all calls currently 
        /// are based on this, but this allows for future expansion. If needed this function can be adapted in this way.
        /// Noteably if this is not used and a function adds a rename profile with the same name as an existing one, the existing one will end
        /// up selected because the selection is based on the profile name alone and not the instance. I expect this is because 
        /// refreshProfilesComboBox() generates a new list each run but I have not looked into this throughly.
        /// </summary>
        /// <param name="text">The text to set.</param>
        private void smartSetProfileNameComboBox(string text)
        {
            if (profileNameComboBox.Text != text)
            {
                profileNameComboBox.Text = text;
            }
        }

        /// <summary>
        /// Updates the values in the CategoriesComboBox based on all possible values from the rename profiles.
        /// </summary>
        private void refreshCategoriesComboBox()
        {
            categoryComboBox.Items.Clear();

            categoryComboBox.Items.AddRange(findCategories(renameProfiles).ToArray());

            if (activeRenameProfile != null)
            {
                smartSetCategoryComboBox(activeRenameProfile.Category);
            }
            else
            {
                smartSetCategoryComboBox("");
            }

            refreshProfilesComboBox();
        }

        /// <summary>
        /// Updates the values in the ProfilesComboBox based on all possible values from the rename profiles that exist in the current category.
        /// </summary>
        private void refreshProfilesComboBox()
        {
            profileNameComboBox.Items.Clear();

            List<RenameProfile> categoryRenameProfiles = findProfilesInCategory(renameProfiles, categoryComboBox.Text);

            profileNameComboBox.Items.AddRange(categoryRenameProfiles.ToArray());

            if (activeRenameProfile != null)
            {
                smartSetProfileNameComboBox(activeRenameProfile.ProfileName);
            }
            else
            {
                smartSetProfileNameComboBox("");
            }

            refreshActiveRenameProfile();
        }

        /// <summary>
        /// Finds and returns all categories names (as a list of strings) that are in the rename profile list. 
        /// </summary>
        /// <param name="rp">The rename profile list to extract categories from.</param>
        /// <returns>A list of all unique category names.</returns>
        private List<string> findCategories(List<RenameProfile> rp)
        {
            List<string> cateogries = new List<string>();

            for (int i = 0; i < rp.Count; i++)
            {
                string category = rp[i].Category;
                if (!cateogries.Contains(category))
                {
                    cateogries.Add(category);
                }
            }

            return cateogries;
        }

        /// <summary>
        /// Finds and returns all profile names (as a list of strings) that are in the rename profile list that belong to the category specified. 
        /// </summary>
        /// <param name="rp">The rename profile list to extract profile names from.</param>
        /// <param name="category">The category that the rename profile must belong to.</param>
        /// <returns>A list of all profile names.</returns>
        private List<RenameProfile> findProfilesInCategory(List<RenameProfile> rp, string category)
        {
            List<RenameProfile> matches = new List<RenameProfile>();

            for (int i = 0; i < rp.Count; i++)
            {
                if (rp[i].Category == category)
                {
                    matches.Add(rp[i]);
                }
            }

            return matches;
        }

        /// <summary>
        /// Generates a relative specific path based on the base path if it contains it.
        /// </summary>
        private void simplifyPath()
        {
            string basePath = baseDirectoryTextBox.Text;
            string path = specificDirectoryTextBox.Text;

            if ((basePath != "") && (path.ToLower().StartsWith(basePath.ToLower())))
            {
                path = RenameProfile.LocalDirectoryChar + path.Substring(basePath.Length);
                specificDirectoryTextBox.Text = path;
            }
        }

        /// <summary>
        /// Updates the activeRenameProfile based on the current profileNameComboBox selection.
        /// </summary>
        private void updateActiveRenameProfile()
        {
            object selection = profileNameComboBox.SelectedItem;

            if (selection != null)
            {
                activeRenameProfile = (RenameProfile)selection;
            }
            else
            {
                setAutoRename(false);
            }

            refreshActiveRenameProfile();
        }

        /// <summary>
        /// Creates a new rename profile based on the information the user has entered into the rename profile section.
        /// </summary>
        private void createNewRenameProfile()
        {
            RenameProfile newProfile = new RenameProfile();

            // Probably Should Vaildate Data Here.

            newProfile.Category = categoryComboBox.Text;
            newProfile.ProfileName = profileNameComboBox.Text;
            newProfile.SpecificDirectory = specificDirectoryTextBox.Text;
            newProfile.FileNameBase = fileNameBaseTextBox.Text;
            newProfile.SequenceMask = sequenceMaskTextBox.Text;
            newProfile.AutoUpdateSequenceMask = autoUpdateCheckBox.Checked;

            renameProfiles.Add(newProfile);

            activeRenameProfile = newProfile;

            refreshCategoriesComboBox();
        }

        /// <summary>
        /// Removes the active rename profile.
        /// </summary>
        private void removeActiveRenameProfile()
        {
            object selection = profileNameComboBox.SelectedItem;

            if (selection != null)
            {
                renameProfiles.Remove((RenameProfile)selection);
                activeRenameProfile = null;
            }

            refreshCategoriesComboBox();
        }

        /// <summary>
        /// Updates the information displayed for the current rename profile based on the information stored in the activeRenameProfile.
        /// </summary>
        private void refreshActiveRenameProfile()
        {
            if (activeRenameProfile != null)
            {
                smartSetCategoryComboBox(activeRenameProfile.Category);
                smartSetProfileNameComboBox(activeRenameProfile.ProfileName);

                specificDirectoryTextBox.Text = activeRenameProfile.SpecificDirectory;
                fileNameBaseTextBox.Text = activeRenameProfile.FileNameBase;
                sequenceMaskTextBox.Text = activeRenameProfile.SequenceMask;
                autoUpdateCheckBox.Checked = activeRenameProfile.AutoUpdateSequenceMask;
            }
            else
            {
                smartSetCategoryComboBox("");
                smartSetProfileNameComboBox("");

                specificDirectoryTextBox.Text = "";
                fileNameBaseTextBox.Text = "";
                sequenceMaskTextBox.Text = "";
                autoUpdateCheckBox.Checked = false;
            }

            updateInfoLabels();
        }

        /// <summary>
        /// Updates the current rename profile based on the information the user has entered into the rename profile section.
        /// </summary>
        private void changeActiveRenameProfileSettings()
        {
            if (activeRenameProfile != null)
            {
                // Probably Should Vaildate Data Here.

                activeRenameProfile.Category = categoryComboBox.Text;
                activeRenameProfile.ProfileName = profileNameComboBox.Text;
                activeRenameProfile.SpecificDirectory = specificDirectoryTextBox.Text;
                activeRenameProfile.FileNameBase = fileNameBaseTextBox.Text;
                activeRenameProfile.SequenceMask = sequenceMaskTextBox.Text;
                activeRenameProfile.AutoUpdateSequenceMask = autoUpdateCheckBox.Checked;

                refreshCategoriesComboBox();
            }
        }

        /// <summary>
        /// Saves the user changeable settings to file.
        /// 
        /// These are saved in two ways, global and rename profile components. All values are stored as category value pairs. 
        /// See a generated file for how the file structure works it is fairly intitive.
        /// </summary>
        /// <returns>Whether the settings were saved successfully (currently dummy as it does not check and the program will crash if their is an error).</returns>
        private bool saveSettings()
        {
            if (allowSaveSettings == true)     // Only allow the settings to be saved if this is the main instance of the program, to stop dummy instances overwriting an older version.
            {
                if (System.IO.File.Exists(settingsPath))
                {
                    DateTime currentModifiedTime = System.IO.File.GetLastWriteTime(settingsPath); // Get the time the file was last written to so that if it is updated we can prevent concurrency issues.

                    if (currentModifiedTime > fileLastModifiedTime)
                    {
                        DialogResult result = MessageBox.Show("The settings file has been updated since the time when the settings file was loaded, this could indicate concurency problems and cause the loss of some settings.\r\n\r\nDo you want to overwrite the settings file?", "Settings Concurrancy", MessageBoxButtons.YesNo);

                        if (result == DialogResult.No)
                        {
                            return false;
                        }
                    }
                }

                System.IO.StreamWriter file = new System.IO.StreamWriter(settingsPath);

                // Write the global settings to file.
                writeValue(file, FileFocusApplication, focusProcessName.ToString());
                writeValue(file, FileConfirmRename, confirmRename.ToString());
                writeValue(file, FileUseThumbnailMode, useThumbnailMode.ToString());
                writeValue(file, FileAutoRename, autoRenameCheckBox.Checked.ToString());
                writeValue(file, FileAutoRenameAllowedAutomated, autoRenameAllowedAutomated.ToString());
                writeValue(file, FileAutoRenameAllowedManual, autoRenameAllowedManual.ToString());
                writeValue(file, FileBaseDirectory, baseDirectoryTextBox.Text);

                // Write all rename profiles to file.
                for (int i = 0; i < renameProfiles.Count; i++)
                {
                    writeProfileDelimiter(file);

                    if ((activeRenameProfile != null) && (activeRenameProfile == renameProfiles[i]))
                    {
                        writeValue(file, FileLastProfile, bool.TrueString);
                    }

                    writeValue(file, FileProfileCategory, renameProfiles[i].Category);
                    writeValue(file, FileProfileName, renameProfiles[i].ProfileName);
                    writeValue(file, FileProfileDirectory, renameProfiles[i].SpecificDirectory);
                    writeValue(file, FileProfileFileName, renameProfiles[i].FileNameBase);
                    writeValue(file, FileProfileMask, renameProfiles[i].SequenceMask);
                    writeValue(file, FileProfileAutoUpdate, renameProfiles[i].AutoUpdateSequenceMask.ToString());
                }

                file.Close();

                fileLastModifiedTime = System.IO.File.GetLastWriteTime(settingsPath); // Update the time the file was last written to so that if it is updated we can prevent concurrency issues.
            }

            return true;
        }

        /// <summary>
        /// Loads the user changeable settings from file.
        /// </summary>
        /// <returns></returns>
        private bool loadSettings()
        {
            System.IO.StreamReader file;
            string line;

            string category = "";
            string value = "";

            bool renameProfileUpdated = false;
            RenameProfile rp = new RenameProfile();

            bool autoRenameSet = false;
            bool autoRename = false;

            fileLastModifiedTime = DateTime.Now;
            if (System.IO.File.Exists(settingsPath))
            {
                fileLastModifiedTime = System.IO.File.GetLastWriteTime(settingsPath); // Get the time the file was last written to so that if it is updated we can prevent concurrency issues.

                renameProfiles.Clear();

                file = new System.IO.StreamReader(settingsPath);
                line = file.ReadLine();

                // While there is text in the file.
                while (line != null)
                {
                    // All lines should be a profile delimeter or a category / value pair.
                    if (isProfileDelimiter(line))
                    {
                        if (renameProfileUpdated)
                        {
                            renameProfiles.Add(rp);
                            rp = new RenameProfile();
                            renameProfileUpdated = false;
                        }
                    }
                    else
                    {
                        // Should be a category / value pair, so try to split it.
                        if (splitValue(line, out category, out value))
                        {
                            // If we split it successfully, find out what the setting means and set it appropriately.
                            if (category.Equals(FileFocusApplication))
                            {
                                setFocusApplication(value);
                            }
                            if (category.Equals(FileConfirmRename))
                            {
                                bool parsedValue;
                                if (bool.TryParse(value, out parsedValue))
                                {
                                    setConfirmRename(parsedValue);
                                }
                            }
                            if (category.Equals(FileUseThumbnailMode))
                            {
                                bool parsedValue;
                                if (bool.TryParse(value, out parsedValue))
                                {
                                    setUseThumbnailMode(parsedValue);
                                }
                            }
                            if (category.Equals(FileAutoRename))
                            {
                                bool parsedValue;
                                if (bool.TryParse(value, out parsedValue))
                                {
                                    autoRenameSet = true;
                                    autoRename = parsedValue;
                                }
                            }
                            if (category.Equals(FileAutoRenameAllowedAutomated))
                            {
                                bool parsedValue;
                                if (bool.TryParse(value, out parsedValue))
                                {
                                    setAutoRenameAllowedAutomated(parsedValue);
                                }
                            }
                            if (category.Equals(FileAutoRenameAllowedManual))
                            {
                                bool parsedValue;
                                if (bool.TryParse(value, out parsedValue))
                                {
                                    setAutoRenameAllowedManual(parsedValue);
                                }
                            }
                            if (category.Equals(FileBaseDirectory))
                            {
                                baseDirectoryTextBox.Text = value;
                            }
                            if (category.Equals(FileLastProfile))
                            {
                                if (value.ToLower() == bool.TrueString.ToLower())
                                {
                                    activeRenameProfile = rp;
                                }
                            }
                            if (category.Equals(FileProfileCategory))
                            {
                                rp.Category = value;
                                renameProfileUpdated = true;
                            }
                            if (category.Equals(FileProfileName))
                            {
                                rp.ProfileName = value;
                                renameProfileUpdated = true;
                            }
                            if (category.Equals(FileProfileDirectory))
                            {
                                rp.SpecificDirectory = value;
                                renameProfileUpdated = true;
                            }
                            if (category.Equals(FileProfileFileName))
                            {
                                rp.FileNameBase = value;
                                renameProfileUpdated = true;
                            }
                            if (category.Equals(FileProfileMask))
                            {
                                rp.SequenceMask = value;
                                renameProfileUpdated = true;
                            }
                            if (category.Equals(FileProfileAutoUpdate))
                            {
                                bool parsedValue;
                                if (bool.TryParse(value, out parsedValue))
                                {
                                    rp.AutoUpdateSequenceMask = parsedValue;
                                }
                                renameProfileUpdated = true;
                            }
                        }
                    }

                    line = file.ReadLine();
                }

                // Add the last rename profile if there was one.
                if (renameProfileUpdated)
                {
                    renameProfiles.Add(rp);
                }

                // Delay setting auto rename to the end untill we have loaded all profiles and hopefully the active rename profile if this is set. Otherwise setting it will be an error, but not a bad one as it will just warn the user.
                if (autoRenameSet)
                {
                    setAutoRename(autoRename);
                }

                file.Close();

                refreshCategoriesComboBox();
            }

            return false;
        }

        /// <summary>
        /// Writes the rename profile delimiter to file, to split between rename profiles.
        /// 
        /// Mirrors isProfileDelimiter().
        /// </summary>
        /// <param name="file">The file to write the line to (assumed already open and in a working state).</param>
        private void writeProfileDelimiter(System.IO.TextWriter file)
        {
            file.WriteLine(FileProfileDelimiter);
        }

        /// <summary>
        /// Checks whether the line is a profile delimiter. 
        /// 
        /// Mirrors writeProfileDelimiter().
        /// </summary>
        /// <param name="line">The line to check.</param>
        /// <returns>Whether the line is a valid profile delimiter.</returns>
        private bool isProfileDelimiter(string line)
        {
            return line.Trim().Equals(FileProfileDelimiter.Trim());
        }

        /// <summary>
        /// Writes a category / value pair to file.
        /// 
        /// Mirrors splitValue()
        /// </summary>
        /// <param name="file">The file to write the line to (assumed already open and in a working state).</param>
        /// <param name="category">The category value.</param>
        /// <param name="value">The value for the specified category.</param>
        private void writeValue(System.IO.TextWriter file, string category, string value)
        {
            file.WriteLine(composeValue(category, value));
        }

        /// <summary>
        /// Generates a category / value line.
        /// 
        /// Mirrors splitValue()
        /// </summary>
        /// <param name="category">The category.</param>
        /// <param name="value">The value to pair with the category.</param>
        /// <returns>The combined category value line.</returns>
        private string composeValue(string category, string value)
        {
            return category + FileCategoryDelimiter + value;
        }

        /// <summary>
        /// Splits a category / value line into their components.
        /// 
        /// Mirrors composeValue()
        /// Note: If the line is invalid, the function will do its best and return the category if possible.
        /// </summary>
        /// <param name="line">The line to split.</param>
        /// <param name="category">The category.</param>
        /// <param name="value">The value paired with the category.</param>
        /// <returns>Whether the line was able to be split successfully.</returns>
        private bool splitValue(string line, out string category, out string value)
        {
            bool found = false;
            int delimiterLocation = line.IndexOf(FileCategoryDelimiter);

            category = "";
            value = "";

            if ((delimiterLocation != -1) && (line != string.Empty))
            {
                category = line.Substring(0, delimiterLocation);

                int valueStart = delimiterLocation + FileCategoryDelimiter.Length;
                int valueLength = line.Length - valueStart;

                if (valueLength > 0)
                {
                    value = line.Substring(valueStart, valueLength); 
                }

                found = true;
            }

            return found;
        }

        /// <summary>
        /// Sends a string message to a process using the WM_COPYDATA message.
        /// 
        /// Mirrors processCopiedData()
        /// Note To Self: Don't use CopyData in managed code in future, a TCP connection would be simpler and nicer.
        /// </summary>
        /// <param name="processName">The name of the process to send the message to.</param>
        /// <param name="messageId">The messageID to use, different messages ID's may be used to mean different things I.E. The type of data (in this case string but need not be) or different context (different information E.G. string-filename, string-processName, etc).</param>
        /// <param name="message">The string message to send.</param>
        /// <returns>Returns whether a process with this name was found and the message was attempted to be sent. (Note: Sending the message may not be recieved correctly.)</returns>
        private bool sendStringMessage(string processName, uint messageId, string message)
        {
            // Note To Self: Don't use CopyData in managed code in future, a TCP connection would be simpler and nicer.
            bool success = false;

            System.Diagnostics.Process[] p = System.Diagnostics.Process.GetProcessesByName(processName);  // Find the process we want to send the message to.

            // Send to all running instance of the program (excep this one).
            for (int i = 0; i < p.Length; i++)
            {
                IntPtr otherProgramHandle = p[i].MainWindowHandle;      // For each instance get the handle to it's main window.

                if ((this.Handle != otherProgramHandle) && (otherProgramHandle != IntPtr.Zero))                 // Make sure this it is not this instance (we don't want to send a message to ourselves).
                {
                    CopyDataStruct copyData = new CopyDataStruct();     // SendMessage: CopyDataStruct for packaging the data to be copied into.

                    //IntPtr ptrHeapText = System.Runtime.InteropServices.Marshal.StringToHGlobalAnsi(message); // Method 1: Allocate space and put the string on the heap.

                    int rawMessageLength = System.Text.Encoding.ASCII.GetByteCount(message);                    // Method 2: Find out how many bytes are needed.
                    byte[] rawMessage = System.Text.Encoding.ASCII.GetBytes(message);                           // Method 2: Convert to a byte arrary.
                    IntPtr ptrHeapText = Marshal.AllocHGlobal(rawMessageLength);                                // Medhod 2: Allocate space on the heap.
                    Marshal.Copy(rawMessage, 0, ptrHeapText, rawMessageLength);                                 // Method 2: Copy the byte array to the heap.

                    copyData.dwData = messageId;                                                                // Message ID.
                    //copyData.cbData = (uint)GlobalSize(ptrHeapText);                                          // Method 1: Size of the message being copied.
                    copyData.cbData = (uint)rawMessageLength;                                                   // Method 2: Size of the message being copied.
                    copyData.lpData = ptrHeapText;                                                              // Pointer to the data to be copied, (pointer our string on the heap).

                    IntPtr ptrHeapCopyData = System.Runtime.InteropServices.Marshal.AllocHGlobal(Marshal.SizeOf(typeof(CopyDataStruct)));   // Allocate more space on the heap to store the copy data struct.
                    System.Runtime.InteropServices.Marshal.StructureToPtr(copyData, ptrHeapCopyData, false);                                // Copy the stuct onto the heap.

                    SendMessage(otherProgramHandle, WM_COPYDATA, this.Handle, ptrHeapCopyData);                                             // Send the message.

                    int error = Marshal.GetLastWin32Error();                                                                                // Get an error code if it failed.

                    if (error == 0)
                    {
                        success = true;
                    }
                    else
                    {
                            #if DEBUG
                            if (System.Diagnostics.Debugger.IsAttached)
                            {
                                MessageBox.Show("Error Sending Message: " + error.ToString());
                            }
                            #endif
                    }

                    Marshal.FreeHGlobal(ptrHeapText);                   // Free the space on the heap we allocated for the string.
                    Marshal.FreeHGlobal(ptrHeapCopyData);               // Free the space on the heap we allocated for the copy data struct.
                }
            }

            return success;
        }

        /// <summary>
        /// Decodes the message in a CopyDataStruct. Currently only decodes a filename passed to load as the active image.
        /// 
        /// Mirrors sendStringMessage()
        /// Note To Self: Don't use CopyData in managed code in future, a TCP connection would be simpler and nicer.
        /// </summary>
        /// <param name="data">The data to use.</param>
        private void processCopiedData(CopyDataStruct data)
        {
            //Note To Self: Don't use CopyData in managed code in future, a TCP connection would be simpler and nicer.
            uint messageId = data.dwData;

            switch (messageId)
            {
				// Process the to cases together untill they are actually loaded.
                case (uint)MessageType.FILE_NAME:
                case (uint)MessageType.BATCH_FILE_NAMES:
                    //string message = System.Runtime.InteropServices.Marshal.PtrToStringAnsi(data.lpData);     // Method 1: Get the string...at the moment we dont care about the size of the data copied or message id, but we could use these in future.

                    int rawMessageLength = (int)data.cbData;                                                    // Method 2: Find the number of bytes / message length.
                    byte[] rawMessage = new byte[rawMessageLength];                                             // Method 2: Create space to store the message.
                    Marshal.Copy(data.lpData, rawMessage, 0, rawMessageLength);                                 // Method 2: Put the message data into raw message.
                    string fileName = Encoding.ASCII.GetString(rawMessage, 0, rawMessageLength);                // Method 2: Convert the raw message to a string.

                    if (messageId == (uint)MessageType.FILE_NAME)
                    {
                        loadImage(fileName, autoRenameAllowedAutomated);
                    }
                    if (messageId == (uint)MessageType.BATCH_FILE_NAMES)
                    {
                        List<string> files = mangleBatchFilenames(fileName);

                        newBatch(files);
                    }
                    break;
                default:
                            #if DEBUG
                            if (System.Diagnostics.Debugger.IsAttached)
                            {
                                MessageBox.Show("Message Id Not Supported: " + messageId);
                            }
                            #endif

                    break;
            }
        }

        /// <summary>
        /// Updates the visual state of the form when batches are queued or processed.
        /// </summary>
        private void updateBatchVisualState()
        {
            if (batchesWaiting())
            {
                loadBatchButton.Visible = true;
            }
            else
            {
                loadBatchButton.Visible = false;
            }
        }

        /// <summary>
        /// Determines whether there are any batches awaiting processing.
        /// </summary>
        /// <returns>Whether any batches are waiting to be processed (true = batch(es) waiting, false = no batches waiting).</returns>
        private bool batchesWaiting()
        {
            return (batches.Count > 0);
        }

        /// <summary>
        /// Called when a new batch is loaded to be processed.
        /// 
        /// This takes care of asking the user whether they want to process the batch now or whether they want to wait and explicitly load them when the
        /// user requests (clicks the batch load button).
        /// </summary>
        /// <param name="filenames">The list of filenames that are to be batch processed.</param>
        private void newBatch(List<string> filenames)
        {
            bool exisitingBatches = batchesWaiting();

            BatchJob b = new BatchJob(filenames);
            batches.Enqueue(b);

            if (!exisitingBatches)
            {
                DialogResult r = MessageBox.Show("A new batch of images has been loaded, are you ready to process these now?", "Batch Process", MessageBoxButtons.YesNo);
                if (r == DialogResult.Yes)
                {
                    loadBatch();
                }
            }

            updateBatchVisualState();
        }

        /// <summary>
        /// Loads the next batch in the queue into ScanHelper. (Actually processes them.)
        /// </summary>
        private void loadBatch()
        {
            BatchJob currentBatch;
            List<String> batchList;

            if (batchesWaiting())
            {
                //DialogResult r = MessageBox.Show("You are about to load a batch of images this may take some time and the application may seem unresponsive. Are you sure you want to continue.", "Batch Warning", MessageBoxButtons.YesNo);
                //if (r == DialogResult.Yes)
                //{
                    if (!AutoRenameActive)
                    {
                        DialogResult r = MessageBox.Show("You are about to load a batch of images but the auto renamer is not on. Do you want to turn it on?", "AutoRenamer State", MessageBoxButtons.YesNo);
                        if (r == DialogResult.Yes)
                        {
                            setAutoRename(true);
                        }
                    }

                    currentBatch = batches.Dequeue();
                    batchList = currentBatch.Filenames;

                    foreach (string filename in batchList)
                    {
                        loadImage(filename, autoRenameAllowedBatch);
                        Application.DoEvents(); // A make the application a little more responsive.
                    }
                //}
            }

            updateBatchVisualState();

            return;
        }

        /// <summary>
        /// Determines whether the list could be a list of file names mangled into a single string.
        /// </summary>
        /// <param name="s"></param>
        /// <returns></returns>
        private bool isMangledFlenames(string s)
        {
            return s.Contains(filenameConjoiner);
        }

        /// <summary>
        /// Turns a list of filenames into a single string of filenames.
        /// 
        /// The list is of the form "filename" "filename" "filename". This means the string can be unmangled even if the filenames contain spaces.
        /// Note the complimentry function is the function overload with different arguments mangleBatchFilenames().
        /// </summary>
        /// <param name="l">A list of filenames.</param>
        /// <returns>The single string containing all of the filenames.</returns>
        private string mangleBatchFilenames(List<string> l)
        {
            string conjoined = "";

            foreach (string s in l)
            {
                if (isMangledFlenames(s))   // Checks if the filename contains the conjoiner and will therefore create problems when we split it.
                {
                    MessageBox.Show("Debug assertation failed. The filename contains " + filenameConjoiner + 
                        ", need to implement character escaping to make this work.\n\n" +
                        "Note: This file will not be processed: \n" + s);
                }
                else
                {
                    conjoined += filenameConjoiner + s + filenameConjoiner + " ";
                }
            }

            return conjoined;
        }

        /// <summary>
        /// Parses a single string of filenames into a list of filenames.
        /// 
        /// Takes a list of the form "filename" "filename" "filename" and parses them into a list of strings.
        /// Note the complimentry function is the function overload with different arguments mangleBatchFilenames() which turns a list of strings into a single string for parsing.
        /// </summary>
        /// <param name="s">The single string containing all of the filenames.</param>
        /// <returns>A list of filenames.</returns>
        private List<string> mangleBatchFilenames(string s)
        {
            List<string> stringList = new List<string>();
            string[] delimiters = new string[1];
            delimiters[0] = filenameConjoiner;
            string tempFilename;

            string[] all = s.Split(delimiters, StringSplitOptions.RemoveEmptyEntries);

            foreach (string filename in all)
            {
                tempFilename = filename.Trim();
                if (tempFilename.Length != 0)
                {
                    stringList.Add(tempFilename);
                }
            }

            return stringList;
        }
    }
}