/**
 * @file RenameProfile.cs
 * @author Yarrago
 * @note Copyright  Yarrago (yarrago.com) 2009 - 2010
 * @version 1.0.0
 * @date Created: 23/11/2009 ?
 * @date Modified: 04/02/2010
 * 
 * @note Modified by Name on --/--/2010. Description:
 * 
 * 
 * 
 * 
 * 
 * License: "Modified BSD License" 3-clause license.
 * 
 * Copyright (c) 2010, Yarrago (yarrago.com)
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without modification, are 
 * permitted provided that the following conditions are met:
 *   - Redistributions of source code must retain the above copyright notice, this list of 
 *     conditions and the following disclaimer.
 *   - Redistributions in binary form must reproduce the above copyright notice, this list of 
 *     conditions and the following disclaimer in the documentation and/or other materials provided
 *     with the distribution.
 *   - Neither the name of the yarrago.com nor the names of its contributors may be used to endorse
 *     or promote products derived from this software without specific prior written permission.
 *   
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR 
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR 
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR 
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR 
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 */

using System;
using System.Collections.Generic;
using System.Text;

namespace ScanHelper
{
    class RenameProfile
    {
        public static char LocalDirectoryChar = System.IO.Path.DirectorySeparatorChar; // Should be const.
        private const int fileNameFindTime = 10;

        private static string commonDirectory = "";

        private string name;                            // The name of the rename profile.
        private string category;                        // The category that the rename profile belongs to.
        private string specificDirectory;               // The specific directory that this rename profile will use.
        private string fileNameBase;                    // The static first base portion of the file name.
        private string sequenceMask;                    // The sequence mask of the file name (used to index a particular file in the sequence).
        private uint sequenceNumber;                    // The next sequence number (typically derived from the sequence mask).
        private bool autoUpdateSequenceNumber;          // Whether the sequence number should be updated after the generated file name containing the sequence number has been used, or should remain constant.
        private bool sequenceNumberUsed = false;        // Whether the sequence number has been used. This provides a safeguard against the sequence number being incremented when it hasn't can't have been used.

        /// <summary>
        /// Gets or sets the common base directory for all profiles.
        /// 
        /// The common directory is prepended to the specific directory if it starts with the LocalDirectoryChar.
        /// </summary>
        public static string CommonDirectory
        {
            get
            {
                return commonDirectory;
            }
            set
            {
                commonDirectory = value;
            }
        }

        /// <summary>
        /// Gets the profile name.
        /// </summary>
        /// <returns>The profile name.</returns>
        public override string ToString()
        {
            return ProfileName;
        }

        /// <summary>
        /// Gets or sets the profiles name.
        /// </summary>
        public string ProfileName
        {
            get
            {
                return name;
            }
            set
            {
                name = value;
            }
        }

        /// <summary>
        /// Gets or sets which cateogry the rename profile belongs to.
        /// </summary>
        public string Category
        {
            get
            {
                return category;
            }
            set
            {
                category = value;
            }
        }
        
        /// <summary>
        /// Gets or sets the rename profiles specific directory.
        /// 
        /// Note the specific directory can be a complete directory or can be a partial directory when it starts with LocalDirectoryChar.
        /// E.g. CommonDirectory = "*Whatever*" SpecificDirectory = "C:\Folder\" ActualDirectory becomes "C:\Folder\"
        /// E.g. CommonDirectory = "C:\" SpecificDirectory = "\Folder\" ActualDirectory becomes "C:\Folder\"
        /// E.g. CommonDirectory = "*Whatever*" SpecificDirectory = "\" ActualDirectory becomes "" - Which puts the file where it was previously.
        /// E.g. CommonDirectory = "*Whatever*" SpecificDirectory = "\\" ActualDirectory becomes "\" - Which puts the file relitive to the application.
        /// </summary>
        public string SpecificDirectory
        {
            get
            {
                return specificDirectory;
            }
            set
            {
                specificDirectory = value;
            }
        }

        /// <summary>
        /// Gets or sets the base filename, which is the first constant unchanging bit of the file name.
        /// </summary>
        public string FileNameBase
        {
            get
            {
                return fileNameBase;
            }
            set
            {
                fileNameBase = value;
            }
        }
        
        /// <summary>
        /// Gets or sets the sequence mask portion of the file name.
        ///         
        /// Contains placeholder digits representing the first sequence number to start with and the locations for the digits to be inserted.
        /// See also the documentation for the sequenceString() function.
        /// </summary>
        public string SequenceMask
        {
            get
            {
                return sequenceMask;
            }
            set
            {
                sequenceMask = value;

                seqenceMaskToSequenceNumber();
                sequenceNumberUsed = false;
            }
        }

        /// <summary>
        /// Sets the sequence number, note typically this should be done by setting the sequence mask, but this function can be used once the mask has been set to set the value to a predefined number.
        /// </summary>
        public uint SequenceNumber
        {
            private get // Private
            {
                return sequenceNumber;
            }
            set // Public
            {
                sequenceNumber = value;
                sequenceNumberUsed = false;

                sequenceMask = sequenceString(sequenceMask, sequenceNumber);
            }
        }

        /// <summary>
        /// Gets or sets whether the next sequence number should be updated next time a filename is generated and used or not.
        /// 
        /// Note: If AutoUpdateSequenceMask is true the sequence mask will be updated to the number that was last returned by 
        /// getNextFileName(), I.E. the last number used. Once the file name has been used (as getNextFileName() may be 
        /// called several times before the path is used) the caller should call filenameUsed(), the sequenceMask will then be 
        /// updated to the next sequence number (which may or may not be already used, but will ensure that the last number 
        /// used is not tried again even if it doesn't exist). If AutoUpdateSequenceMask is false getNextFileName() will use the
        /// next avaliable number after the sequence number (checking which numbers already exist and use the next avaliable one).
        /// </summary>
        public bool AutoUpdateSequenceMask
        {
            get
            {
                return autoUpdateSequenceNumber;
            }
            set
            {
                autoUpdateSequenceNumber = value;
            }
        }

        /// <summary>
        /// Whether the rename profile is valid.
        /// </summary>
        public bool Valid
        {
            get
            {
                return validatePath(genteratePath(sequenceNumber, ""));
            }
        }

        /// <summary>
        /// Converts the stored sequence mask into the sequence number.
        /// 
        /// Converts the number in the sequence mask to the starting number for the sequence series.
        /// EG. sequenceMask = "STRING0001" sequenceNumber =    1
        /// EG. sequenceMask = "STRING0003" sequenceNumber =    3
        /// EG. sequenceMask = "STRING3003" sequenceNumber = 3003
        /// EG. sequenceMask = "S1T2R3I4NG" sequenceNumber = 1234
        /// Note if no digits were converted the sequence number is set to zero.
        /// Result: If the sequence mask contains placeholders for digits to be converted and the mask was able to be converted into an inital sequence number.
        /// </summary>
        /// <returns>Whether the conversion was successful.</returns>
        private bool seqenceMaskToSequenceNumber()
        {
            bool converted = false;

            sequenceNumber = 0;
            for (int i = 0; i < sequenceMask.Length; i++)
            {
                char c = sequenceMask[i];
                if (char.IsDigit(c))
                {
                    converted = true;
                    sequenceNumber = sequenceNumber * 10 + uint.Parse(c.ToString());
                }
            }

            return converted;
        }

        /// <summary>
        /// Generates a new sequence string using the old sequence string and sequence number.
        /// </summary>
        /// <returns></returns>
        private string currentSequenceString()
        {
            return sequenceString(sequenceMask, sequenceNumber);
        }

        /// <summary>
        /// Generates a sequence mask, updating the digits in the mask using the number in the sequenceCount.
        /// 
        /// Replaces characters from the sequence count into the digit holders of the sequence mask.
        /// EG. mask = "STRING0001" sequenceCount =     1 Result = "STRING0001"
        /// EG. mask = "STRING0001" sequenceCount =     3 Result = "STRING0003"
        /// EG. mask = "STRING0001" sequenceCount = 30003 Result = "STRING0003"
        /// EG. mask = "STRING0001" sequenceCount =   303 Result = "STRING0303"
        /// EG. mask = "S0T0R0I1NG" sequenceCount =  1234 Result = "S1T2R3I4NG"
        /// Future Idea: Add '*' to string to represent any additional numbers. 
        /// EG. mask = "STRING*"     sequenceCount =      3 Result = "STRING3"
        /// EG. mask = "STRING*1"    sequenceCount =      3 Result = "STRING3"
        /// EG. mask = "STRING*0001" sequenceCount =      3 Result = "STRING0003"
        /// EG. mask = "STRING*0001" sequenceCount = 120003 Result = "STRING120003"
        /// </summary>
        /// <param name="mask">The sequence mask to use as a basis for the new sequence string.</param>
        /// <param name="sequenceCount">The number to put into the new mask.</param>
        /// <returns>The new sequence mask.</returns>
        private string sequenceString(string mask, long sequenceCount)
        {
            string sequenceCountString = sequenceCount.ToString();  // Variable to hold the part of the sequence count that hasn't been converted to the sequence string.
            string sequenceString = mask;                           // Variable to hold the final sequence string.

            // Replace digits with sequence characters from right to left.
            for (int i = sequenceString.Length - 1; i >= 0; i--)
            {
                if (char.IsDigit(sequenceString[i]))
                {
                    // Replace a digit from the sequence count into the next digit from the mask.
                    // This is really crap way of replacing a character but is easy.
                    char[] replace = sequenceString.ToCharArray();
                    char c = '0';
                    if (sequenceCountString.Length >= 1)
                    {
                        c = sequenceCountString[sequenceCountString.Length - 1];
                        sequenceCountString = sequenceCountString.Substring(0, sequenceCountString.Length - 1); // Remove the least significant digit from the string.
                    }
                    replace[i] = c;
                    sequenceString = new string(replace);    // Take the least significant digit from the count and add it to the sequence string.
                }
            }

            return sequenceString;
        }

        /// <summary>
        /// Determines whether the path supplied is a vaild path and can be used.
        /// 
        /// Note: For now we assume the path is valid, but in the future this could be updated to validate the path.
        /// </summary>
        /// <param name="path">The path to validate.</param>
        /// <returns>Whether the path is valid.</returns>
        private bool validatePath(string path)
        {
            return true;
        }
        
        /// <summary>
        /// Generates a path based on the rename profile, using the specific sequence number and extension.
        /// 
        /// </summary>
        /// <param name="sequence">The sequence number to use when generating the path.</param>
        /// <param name="extension">The extension to use for the generated file name.</param>
        /// <returns>A full path and filename including extension.</returns>
        private string genteratePath(long sequence, string extension)
        {
            /// Note: Could have used System.IO.Path.Combine, but this works for now and so don't want to change it an possibly break it.
            
            string dir = "";

            if (specificDirectory.Length > 0)
            {
                // If the first character is not the local directory character the directory must be a global directory.
                if (specificDirectory[0] != LocalDirectoryChar)
                {
                    dir = specificDirectory;
                }
                else
                {
                    // Start with the commonDirectory.
                    dir = commonDirectory;

                    // If the last character of the common directory is not the directory separater (I.E. '/') append it.
                    dir = endWithDirectorySeparatorChar(dir);

                    // Add the specific directory excluding the LocalDirectoryChar (which is the first character).
                    dir += specificDirectory.Substring(1);
                }
            }

            // If the last directory char is not the directory seperator char appended it so the file name will follow. 
            dir = endWithDirectorySeparatorChar(dir);

            return dir + fileNameBase + sequenceString(sequenceMask, sequence) + extension;
        }

        /// <summary>
        /// Makes sure the directory ends with a directory separator character (adds a single character if necessary, does not alter otherwise).
        /// 
        /// Makes sure the returned string will end with a directory seperator char.
        /// Input: "C:\Folder" Output "C:\Folder\"
        /// Input: "C:\Folder\" Output "C:\Folder\"
        /// </summary>
        /// <param name="dir">The path to check.</param>
        /// <returns>The path with a directory separator char appended if required.</returns>
        private string endWithDirectorySeparatorChar(string dir)
        {
            if (dir.Length > 0)
            {
                char lastChar = dir[dir.Length - 1];
                if ((lastChar != System.IO.Path.DirectorySeparatorChar) && (lastChar != System.IO.Path.AltDirectorySeparatorChar))
                {
                    dir += System.IO.Path.DirectorySeparatorChar;
                }
            }

            return dir;
        }

        /// <summary>
        /// Generates the next filename that is not used based on the rename profile.
        /// 
        /// Note: The main loop could loop forever if the mask contains no digits and the base file exists.
        /// The time check has been added as a stop gap measure to force failure after a number of seconds if a suitable name is not found.
        /// </summary>
        /// <param name="extension">The extension to use for the generated file.</param>
        /// <returns>The generated file name. ("" on failure.)</returns>
        public string getNextFileName(string extension)
        {
            uint nextSequence = sequenceNumber;
            string lastFileName = "";              
            string nextFileName = "";
            bool found = false;

            if (Valid)
            {
                System.DateTime start = System.DateTime.Now;

                found = false;
                while ((!found) && (start.AddSeconds(fileNameFindTime) >= System.DateTime.Now))
                {
                    lastFileName = nextFileName;
                    nextFileName = genteratePath(nextSequence, extension);
                    if (validatePath(nextFileName))
                    {
                        // Note: Last file name is just used to check if the file names are all the same and if they are it just returns this irrespective of whether it exists or not.
                        if (lastFileName == nextFileName)
                        {
                            found = true;
                            break;
                        }

                        if (!System.IO.File.Exists(nextFileName))
                        {
                            // Can use this filename.
                            found = true;
                        }
                        else
                        {
                            nextSequence++;
                        }
                    }
                    else
                    {
#if DEBUG
                        if (System.Diagnostics.Debugger.IsAttached)
                        {
                            System.Windows.Forms.MessageBox.Show("Error: RenameProfile - SequenceNumber Invalidates Path.\r\n\r\nPath Is: " + nextFileName);
                        }
#endif
                        // Strange error...the number has made it invalid.
                        // Don't do anything about this now, except fail.
                        break;
                    }
                }
            }
           
            if (found == true)   
            {
                if (autoUpdateSequenceNumber)
                {
                    // We auto update the sequence number to the found sequence number.
                    // Note: We have to wait till it is confirmed that this sequence number has actually been used befor we go to the next one.
                    //       That is achieved by calling the fileName used function. This does place some dependancy on the user to make sure this
                    //       happens but we have no other way of identifying the name has actually been consumed and used. If the do forget to call
                    //       this then autoUpdate sequence number still works, but just updates to the last found instead of the next tried sequence number.
                    SequenceNumber = nextSequence;
                    sequenceNumberUsed = true;
                }

                return nextFileName;   
            }

            return ""; // Error
        }

        /// <summary>
        /// The last filename (and therefore sequence number) have been used...AKA we want to auto update to the next one if AutoUpdate is on.
        /// </summary>
        public void filenameUsed()
        {
            // sequenceNumberUsed provides a safegaurd against multiple calls to this function without generating new filenames between (hence they can't have been used).
            if ((autoUpdateSequenceNumber) && (sequenceNumberUsed == true))
            {
                sequenceNumberUsed = false;
                SequenceNumber++;
            }
        }
    }
}
